#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------------
# @authors   Christoph Dinh <christoph.dinh@brain-link.de>
#            John G Samuelson <johnsam@mit.edu>
# @version   1.0
# @date      September, 2017
# @copyright Copyright (c) 2017-2022, authors of CMNE. All rights reserved.
# @license   MIT
# @brief     Settings
# ---------------------------------------------------------------------------

#%%
import os

###################################################################################################
# Settings class
###################################################################################################

class Settings(object):
    """the settings object

    Attributes:
        _data_path: Path to the MEG data.
        _results_path: Result path
        _fname_raw: Raw file
        _fname_inv: Inverse file
        _fname_event: Event file
        _fname_test_idcs: Test indeces file
        _tb_log_dir: Tensor board directory.
        _modality: The selected modality meg or meg-eeg
    """
    
    _large_memory = True # If all epoch data fit into the Ram
    
    ###############################################################################################
    # Constructor
    ###############################################################################################
    def __init__(self, results_path='/workspace/results/', data_path='/workspace/data/', 
                 fname_raw='assr_270LP_fs900_raw.fif',
                 fname_inv='assr_270LP_fs900_raw-ico-4-meg-eeg-inv.fif',
                 fname_eve='assr_270LP_fs900_raw-eve.fif',
                 fname_test_idcs='assr_270LP_fs900_raw-test-idcs.txt',
                 meg_and_eeg=True):
        """Return a new Settings object."""
        
        self._meg_and_eeg = meg_and_eeg
        if meg_and_eeg:
            self._modality = 'meg-eeg'
        else: 
            self._modality = 'meg'
        
        self._results_path = results_path
        self._data_path = data_path
        
        self._fname_raw = self._data_path + fname_raw
        self._fname_inv = self._data_path + fname_inv
        self._fname_event = self._data_path + fname_eve
        self._fname_test_idcs = self._data_path + fname_test_idcs
        
        # CMNE
        if not os.path.isdir(self._results_path + 'cmne'):
            os.mkdir(self._results_path + 'cmne') 
        self._results_cmne_path = self._results_path + 'cmne'

        if not os.path.isdir(self._results_path + 'cmne/models'):
            os.mkdir(self._results_path + 'cmne/models')
        self._results_models_path = self._results_path + 'cmne/models'

        if not os.path.isdir(self._results_path + 'cmne/logs'):
            os.mkdir(self._results_path + 'cmne/logs')
        self._results_logs_path = self._results_path + 'cmne/logs'
        self._tb_log_dir = self._results_path + 'cmne/logs'

        if not os.path.isdir(self._results_path + 'cmne/training'):
            os.mkdir(self._results_path + 'cmne/training')
        self._results_training_path = self._results_path + 'cmne/training'

        if not os.path.isdir(self._results_path + 'cmne/img'):
            os.mkdir(self._results_path + 'cmne/img')
        self._results_img_path = self._results_path + 'cmne/img'
            
        # source_estimation
        if not os.path.isdir(self._results_path + 'source_estimations'):    
            os.mkdir(self._results_path + 'source_estimations')
        self._results_stc_path = self._results_path + 'source_estimations'
                    
    
    
    ###############################################################################################
    # Getters and setters
    ###############################################################################################
    def large_memory(self):
        """
        Returns whether large memory configuration is choosen
        """
        return self._large_memory
    
    def meg_and_eeg(self):
        """
        Returns the selected modality
        """
        return self._meg_and_eeg
        
    def modality(self):
        """
        Returns the selected modality
        """
        return self._modality
    
    def data_path(self):
        """
        Returns the data path
        """
        return self._data_path
        
    def results_path(self):
        """
        Returns the repository path
        """
        return self._results_path
            
    def fname_raw(self):
        """
        Returns the raw file name
        """
        return self._fname_raw
        
    def fname_inv(self):
        """
        Returns the inverse operator file name
        """
        return self._fname_inv
    
    def fname_event(self):
        """
        Returns the event file name
        """
        return self._fname_event
    
    def fname_test_idcs(self):
        """
        Returns the file name containing the test indeces
        """
        return self._fname_test_idcs
    
    def tb_log_dir(self):
        """
        Returns the tensor board log dir
        """
        return self._tb_log_dir
    
    def results_cmne_path(self):
        """
        Returns the cmne results path
        """
        return self._results_cmne_path
    
    def results_models_path(self):
        """
        Returns the models results path
        """
        return self._results_models_path

    def results_logs_path(self):
        """
        Returns the logs results path
        """
        return self._results_logs_path

    def results_training_path(self):
        """
        Returns the training results path
        """
        return self._results_training_path

    def results_img_path(self):
        """
        Returns the img results path
        """
        return self._results_img_path

    def results_stc_path(self):
        """
        Returns the stc results path
        """
        return self._results_stc_path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------------
# @authors   Christoph Dinh <christoph.dinh@brain-link.de>
#            John G Samuelson <johnsam@mit.edu>
# @version   1.0
# @date      September, 2017
# @copyright Copyright (c) 2017-2022, authors of CMNE. All rights reserved.
# @license   MIT
# @brief     Data related types and methods, e.g., data loader
# ---------------------------------------------------------------------------

#%%
import os
import numpy as np
import random
import mne
from mne.minimum_norm import apply_inverse_epochs, read_inverse_operator

from .settings import Settings


###################################################################################################
# Standardize
###################################################################################################

def standardize(mat, mean=None, std=None):
	"""
	0 center and scale data
	Standardize an np.array to the array mean and standard deviation or specified parameters
	See https://en.wikipedia.org/wiki/Feature_scaling
	"""
	if mean is None:
		mean = np.mean(mat, axis=1)
	
	if std is None:
		std = np.std(mat, axis=1)

	data_normalized = (mat.transpose() - mean).transpose()
	data = (data_normalized.transpose() / std).transpose()

	return data


###################################################################################################
# Reshape with lookback value for LSTM
###################################################################################################

# creates lookback values using helper, then reshapes for LSTM
def reshape_stc_data(stc, look_back, start=0, step_size=1):
    
    samples,n = stc.shape

    stop = samples - look_back - 1

    feature_parts, label_parts = [], []

    for i in range(start,stop,step_size):
        feature_parts.append(stc[i:(i + look_back), :])
        # Take just the last estimate as label
        label_parts.append(stc[i + look_back, :])

    return feature_parts, label_parts


###################################################################################################
# Reshape with lookback value for LSTM
###################################################################################################

# creates lookback values using helper, then reshapes for LSTM
def reshape_epoch_stc_data(epoch, stc, look_back):
    
    samples_epoch,n = epoch.shape
    samples_stc,n = stc.shape

    epoch_parts, stc_parts = [], []

    # Make sure that samples are equaly long
    if samples_epoch != samples_stc:
        return epoch_parts, stc_parts

    for i in range(samples_epoch - look_back):
        epoch_parts.append(epoch[i:(i + look_back), :])
        # Take the whole estimate sequence as label
        #stc_parts.append(stc[i:(i + look_back), :])
        # Take just the last estimate as label
        stc_parts.append(stc[i + look_back - 1, :])

    return epoch_parts, stc_parts


###################################################################################################
# Reshape with lookback and look into future value for LSTM
###################################################################################################

def reshape_future_data(stc, look_back, future_steps=1, start=0, step_size=1):
    
    samples,n = stc.shape

    stop = samples - look_back - 1 - future_steps + 1

    feature_parts, label_parts = [], []

    for i in range(start,stop,step_size):
        feature_parts.append(stc[i:(i + look_back), :])
        # Take just the last estimate as label
        label_parts.append(stc[i + look_back : i + look_back + future_steps:1, :].flatten())

    return feature_parts, label_parts


###################################################################################################
# Create Sequence Parts
###################################################################################################

def create_sequence_parts(stc, look_back, start=0, step_size=1):
    samples,n = stc.shape

    stop = samples - look_back - 1

    stc_sequence_parts = []

    for i in range(start,stop,step_size):
        stc_sequence_parts.append(stc[i:(i + look_back), :])

    stc_sequences = np.array(stc_sequence_parts)

    return stc_sequences


###################################################################################################
# Data class
###################################################################################################

class Data(object):
    """the cmne data object

    Attributes:
        _settings: CMNE settings object.
        _inv_op: The loaded inverse operator.
        _epochs: The loaded epochs.
        _num_epochs: Number of available epochs.
        _train_idcs: Indeces which should be used for training
        _test_idcs: Indeces which should be used for testing
    """

    _mag_th = 4e-12 #Simulation: 4e-11
    _grad_th = 4000e-13 #Simulation: 4000e-12
    # Using the same inverse operator when inspecting single trials Vs. evoked
    _snr = 3.0  # Standard assumption for average data but using it for single trial
    _lambda2 = 1.0 / _snr ** 2
    _method = "dSPM"
    
	###############################################################################################
	# Constructor
	###############################################################################################
    def __init__(self, settings: Settings):
        """Return a new CMNEData object."""
        self._settings = settings

	###############################################################################################
	# Load Data
	###############################################################################################
    def load_data(self, event_id=1, tmin=-0.2, tmax=0.5, train_percentage = 0.85):
		# Load data
        inverse_operator = read_inverse_operator(self._settings.fname_inv())
        raw = mne.io.read_raw_fif(self._settings.fname_raw())
        events = mne.read_events(self._settings.fname_event())
        
        # Set up pick list
        include = []
        
        # set EEG average reference
#        raw.set_eeg_reference()
        
        # Add a bad channel
        #    raw.info['bads'] += ['EEG 053']  # bads + 1 more
        
        # pick MEG channels
        picks = mne.pick_types( raw.info, meg=True, eeg=self._settings.meg_and_eeg(), stim=False, eog=False, include=include, exclude='bads')
        
        # Read epochs
        epochs = mne.Epochs( raw, events, event_id, tmin, tmax, baseline=(None, 0),
                            picks=picks, preload=self._settings.large_memory(), reject=dict(mag=self._mag_th, grad=self._grad_th))#, eog=150e-5))#eog=150e-6))
        
        epochs.drop_bad()
        
        self._inv_op = inverse_operator
        self._epochs = epochs
        
        #Count epochs - since they are not preloaded it has to be done with a for loop
        num_epochs = 0
        for epoch in epochs:
            num_epochs = num_epochs + 1
        self._num_epochs = num_epochs
        
        whole_list = list(range(num_epochs))
        
        if os.path.isfile(self._settings.fname_test_idcs()):
            self._test_idcs = []
            with open(self._settings.fname_test_idcs(), "r") as f:
              for line in f:
                self._test_idcs.append(int(line.strip()))
                
            self._train_idcs = [item for item in whole_list if item not in self._test_idcs]
        else:
            #split train and test
            random.seed(42)
            self._train_idcs = random.sample(range(num_epochs), (int)(num_epochs*train_percentage))
            self._test_idcs = [item for item in whole_list if item not in self._train_idcs]
            with open(self._settings.fname_test_idcs(), "w") as f:
                for idx in self._test_idcs:
                    f.write(str(idx) +"\n")

    ###############################################################################################
    # Generate Input
    ###############################################################################################
    
    def generate_normalized_input(self, look_back=40, batch_size=20):
        """
        Create the normalized input
        """
        nave = 1

        #Count epochs
        num_epochs = 0
        for epoch in self._epochs:
            num_epochs = num_epochs + 1

        while True:
            # select random epochs
            idx = random.sample(range(num_epochs), batch_size)
            sel_epochs = self.epochs[idx]

            stcs = apply_inverse_epochs(sel_epochs, self._inv_op, self._lambda2, self._method, pick_ori="normal", nave=nave)

            # Attention - just an approximation, since not all stc are considered for the mean and the std
            stc_data = np.hstack([stc.data for stc in stcs])
            stc_mean = np.mean(stc_data, axis=1)
            stc_std = np.std(stc_data, axis=1)
            stc_data = None
            #Attention end

            for stc in stcs:
                stc_normalized = standardize(stc.data, mean=stc_mean, std=stc_std)
                stc_normalized_T = stc_normalized.transpose()
				
                feature_list, label_list = reshape_stc_data(stc = stc_normalized_T, look_back = look_back)

                features = np.array(feature_list)
                labels = np.array(label_list)
                
                yield features, labels
		
    ###############################################################################################
    # Getters and setters
    ###############################################################################################
    def epochs(self, idx=None):
        """
        Returns selected epochs, if selection is None then all epochs are returned
        """
        if idx == None:
            return self._epochs
        else:
            return self._epochs[idx]
            
    def test_idcs(self):
        """
        Returns selected test indeces
        """
        return self._test_idcs
            
    def test_epochs(self, idx=None):
        """
        Returns selected test epochs, if selection is None then all test epochs are returned
        """
        if idx == None:
            return self._epochs[self._test_idcs]
        else:
            return self._epochs[self._test_idcs][idx]

    def train_idcs(self):
        """
        Returns selected test indeces
        """
        return self._train_idcs
    
    def train_epochs(self, idx=None):
        """
        Returns selected test epochs, if selection is None then all test epochs are returned
        """
        if idx == None:
            return self._epochs[self._train_idcs]
        else:
            return self._epochs[self._train_idcs][idx]
            
    def inv_op(self):
        """
        Returns the loaded inverse operator
        """
        return self._inv_op
    
    def method(self):
        """
        Returns the inverse method
        """
        return self._method
    
    def snr(self):
        """
        Returns the datas snr
        """
        return self._snr
    
    def lambda2(self):
        """
        Returns the datas snr
        """
        return self._lambda2
        
    def num_epochs(self):
        """
        Returns the datas snr
        """
        return self._num_epochs
        

###################################################################################################
# Create all data at once
###################################################################################################

def create_lstm_data(epochs, inverse_operator, lambda2, method, look_back = 1):
    """
    Create the dataset for testing regression models in the CNTK format
    Y = GQ + E -> features = stc, labels = stc
    """
    nave = 2
    
    # Compute inverse solution and stcs for each epoch
    # Use the same inverse operator as with evoked data (i.e., set nave)
    # If you use a different nave, dSPM just scales by a factor sqrt(nave)
    stcs = apply_inverse_epochs(epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave)

    ###############################################################################
    # Standardize Label and Mean Data

    feature_data = np.hstack([stc.data for stc in stcs])
    feature_mean = np.mean(feature_data, axis=1)
    feature_std = np.std(feature_data, axis=1)

    features_normalized = []
    labels_normalized = []
    
    for stc in stcs:
        stc_normalized = standardize(stc.data,mean=feature_mean,std=feature_std)
        stc_normalized_T = stc_normalized.transpose()
        
        feature_parts, label_parts = reshape_stc_data(stc = stc_normalized_T, look_back = look_back)

        features_normalized.extend(feature_parts)
        labels_normalized.extend(label_parts)

    features= np.array(features_normalized)
    labels = np.array(labels_normalized)

    return features, labels        


def create_epoch_stc_data(epochs, inverse_operator, lambda2, method, look_back = 1):
    """
    Create the dataset for testing regression models in the CNTK format
    Y = GQ + E
    """
    nave = 2
    
    # Compute inverse solution and stcs for each epoch
    # Use the same inverse operator as with evoked data (i.e., set nave)
    # If you use a different nave, dSPM just scales by a factor sqrt(nave)
    stcs = apply_inverse_epochs(epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave)

    ###############################################################################
    # Standardize Label and Mean Data

    label_data = np.hstack([stc.data for stc in stcs])
    label_mean = np.mean(label_data, axis=1)
    label_std = np.std(label_data, axis=1)

    feature_data = np.hstack([epoch.data for epoch in epochs])
    feature_mean = np.mean(feature_data, axis=1)
    feature_std = np.std(feature_data, axis=1)

    epochs_normalized = []
    stcs_normalized = []
    
    for epoch,stc in zip(epochs,stcs):
        stc_normalized = standardize(stc.data,mean=label_mean,std=label_std)
        stc_normalized_T = stc_normalized.transpose()
        
        epoch_normalized = standardize(epoch,mean=feature_mean,std=feature_std)
        epoch_normalized_T = epoch_normalized.transpose()

        epoch_parts, stc_parts = reshape_epoch_stc_data(epoch = epoch_normalized_T, stc = stc_normalized_T, look_back = look_back)

        epochs_normalized.extend(epoch_parts)
        stcs_normalized.extend(stc_parts)

    features = np.array(epochs_normalized)
    labels = np.array(stcs_normalized)

    return features, labels

      
###################################################################################################
# Generate DNN Batches
###################################################################################################

def generate_dnn_batches(epochs, inverse_operator, lambda2, method, batch_size=20):    
    """    
    Create the DNN Training Batches     
    """    
    nave = 1
    
    # Compute inverse solution and stcs for each epoch    
    # Use the same inverse operator as with evoked data (i.e., set nave)    
    # If you use a different nave, dSPM just scales by a factor sqrt(nave)    
    stcs = apply_inverse_epochs(epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave, verbose=None)
    
    #Count epochs    
    num_epochs = 0    
    for epoch in epochs:    
        num_epochs = num_epochs + 1

    while True:    
        # select random epochs    
        idx = random.sample(range(num_epochs), batch_size)    
        sel_epochs = epochs[idx]    
        sel_stcs = [stcs[i] for i in idx]

        ###############################################################################    
        # Standardize Label and Mean Data

        feature_data = np.hstack([epoch.data for epoch in sel_epochs])    
        feature_mean = np.mean(feature_data, axis=1)    
        feature_std = np.std(feature_data, axis=1)

        label_data = np.hstack([stc.data for stc in sel_stcs])    
        label_mean = np.mean(label_data, axis=1)    
        label_std = np.std(label_data, axis=1)

        features_normalized = standardize(feature_data, mean=feature_mean, std=feature_std)    
        features = features_normalized.transpose()
        
        labels_normalized = standardize(label_data, mean=label_mean, std=label_std)    
        labels = labels_normalized.transpose()
        
        yield features, labels
        
        
###################################################################################################
# Generate DNN Eval Batches
###################################################################################################

def generate_dnn_eval_batches(epochs, inverse_operator, lambda2, method):    
    """    
    Create the DNN Evaluation Batches     
    """    
    nave = 1        
    
    # Compute inverse solution and stcs for each epoch    
    # Use the same inverse operator as with evoked data (i.e., set nave)    
    # If you use a different nave, dSPM just scales by a factor sqrt(nave)    
    stcs = apply_inverse_epochs(epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave, verbose=None)
        
    ###############################################################################    
    # Standardize Label and Mean Data    
    feature_data = np.hstack([epoch.data for epoch in epochs])    
    feature_mean = np.mean(feature_data, axis=1)    
    feature_std = np.std(feature_data, axis=1)
        
    label_data = np.hstack([stc.data for stc in stcs])    
    label_mean = np.mean(label_data, axis=1)    
    label_std = np.std(label_data, axis=1)
        
    features_normalized = standardize(feature_data, mean=feature_mean, std=feature_std)    
    features = features_normalized.transpose()
    
    labels_normalized = standardize(label_data, mean=label_mean, std=label_std)    
    labels = labels_normalized.transpose()
    
    total_num_samples = features.shape[1]
    epoch_num_samples = stcs[0].shape[1]
    # for i in range(0, total_num_samples, epoch_num_samples):
    i = 0
    while True:
        yield features[i:i+epoch_num_samples:1], labels[i:i+epoch_num_samples:1]
        i = i + epoch_num_samples
        i = i % total_num_samples


###################################################################################################
# Generate Future Batches
###################################################################################################

def generate_lstm_batches(epochs, inverse_operator, lambda2, method, look_back=40, batch_size=20):
    """
    Create the LSTM Training Batches 
    """
    nave = 2

    #Count epochs
    num_epochs = 0
    for epoch in epochs:
        num_epochs = num_epochs + 1
    #take 0.75 of the epochs for training
    train_max = num_epochs

    while True:
        # select random epochs
        #idx = random.sample(range(train_max), batch_size)
        idx = np.random.randint(train_max, size=batch_size)
        sel_epochs = epochs[idx]
        # Compute inverse solution and stcs for each epoch
        # Use the same inverse operator as with evoked data (i.e., set nave)
        # If you use a different nave, dSPM just scales by a factor sqrt(nave)
        #sel_epochs = mne.set_eeg_reference(sel_epochs, ref_channels=None, copy=True)[0]
        sel_epochs.apply_proj()

        # Compute inverse solution and stcs for each epoch
        # Use the same inverse operator as with evoked data (i.e., set nave)
        # If you use a different nave, dSPM just scales by a factor sqrt(nave)
        stcs = apply_inverse_epochs(sel_epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave)

        # Attention - just an approximation, since not all stc are considered for the mean and the std
        stc_data = np.hstack([stc.data for stc in stcs])
        stc_mean = np.mean(stc_data, axis=1)
        stc_std = np.std(stc_data, axis=1)
        stc_data = None
        #Attention end

        for stc in stcs:
            stc_normalized = standardize(stc.data,mean=stc_mean,std=stc_std)
            stc_normalized_T = stc_normalized.transpose()
            
            feature_list, label_list = reshape_stc_data(stc = stc_normalized_T, look_back = look_back)

            features = np.array(feature_list)
            labels = np.array(label_list)

            yield features, labels


###################################################################################################
# Generate LSTM Eval Batches
###################################################################################################

def generate_lstm_eval_batches(epochs, inverse_operator, lambda2, method, look_back=40):
    """
    Create the LSTM Training Batches 
    """
    nave = 1

    #Count epochs
    num_epochs = 0
    for epoch in epochs:
        num_epochs = num_epochs + 1

    # # select subset of epochs since numpy std ran into MemoryError
    # if num_epochs > 100:
    #     idx = random.sample(range(num_epochs), 100)
    #     sel_epochs = epochs[idx]
    #     sel_stcs = apply_inverse_epochs(sel_epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave)
    stcs = apply_inverse_epochs(epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave)

    # # select subset of epochs since numpy std ran into MemoryError
    # if num_epochs > 100:
    #     # Attention - just an approximation, since not all stc are considered for the mean and the std
    #     stc_data = np.hstack([stc.data for stc in sel_stcs])
    #     #Attention end
    # else:
    stc_data = np.hstack([stc.data for stc in stcs])
    stc_mean = np.mean(stc_data, axis=1)
    stc_std = np.std(stc_data, axis=1)
    stc_data = None

    count = 0
    while True:
        for stc in stcs:
            stc_normalized = standardize(stc.data,mean=stc_mean,std=stc_std)
            stc_normalized_T = stc_normalized.transpose()
            
            feature_list, label_list = reshape_stc_data(stc = stc_normalized_T, look_back = look_back)

            features = np.array(feature_list)
            labels = np.array(label_list)

            print(">>>> LSTM Eval iteration ",count)

            count = count + 1

            yield features, labels


###################################################################################################
# Generate LSTM Future Batches
###################################################################################################

def generate_lstm_future_batches(epochs, inverse_operator, lambda2, method, look_back=40, future_steps=1, batch_size=20, loop=True):
    """
    Create the Batches
    """

    #%%    epochs.drop(range(3,len(epochs)))#debug purpose keep only the first 3 epochs
    nave = 2 #len(epochs)

    #Count epochs
    num_epochs = 0
    for epoch in epochs:
        num_epochs = num_epochs + 1
    #take 0.75 of the epochs for training
    train_max = num_epochs

    #%%
    condition = True
    while condition:
        if loop == False:
            condition = False

        #%% select random epochs
        idx = np.random.randint(train_max, size=batch_size)
        sel_epochs = epochs[idx]
        
        # Compute inverse solution and stcs for each epoch
        # Use the same inverse operator as with evoked data (i.e., set nave)
        # If you use a different nave, dSPM just scales by a factor sqrt(nave)
        #sel_epochs = mne.set_eeg_reference(sel_epochs, ref_channels=None, copy=True)[0]
        sel_epochs.apply_proj()

        # Compute inverse solution and stcs for each epoch
        # Use the same inverse operator as with evoked data (i.e., set nave)
        # If you use a different nave, dSPM just scales by a factor sqrt(nave)
        stcs = apply_inverse_epochs(sel_epochs, inverse_operator, lambda2, method, pick_ori="normal", nave=nave)

        # Attention - just an approximation, since not all stc are considered for the mean and the std
        stc_data = np.hstack([stc.data for stc in stcs])
        stc_mean = np.mean(stc_data, axis=1)
        stc_std = np.std(stc_data, axis=1)
        stc_data = None
        #Attention end
        #%%
        for stc in stcs:
            #%%
            stc_normalized = standardize(stc.data,mean=stc_mean,std=stc_std)
            stc_normalized_T = stc_normalized.transpose()
            
            #%%
            feature_list, label_list = reshape_future_data(stc=stc_normalized_T, look_back=look_back, future_steps=future_steps)

            features = np.array(feature_list)
            labels = np.array(label_list)

            yield features, labels



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------------
# @authors   Christoph Dinh <christoph.dinh@brain-link.de>
#            John G Samuelson <johnsam@mit.edu>
# @version   1.0
# @date      April, 2018
# @copyright Copyright (c) 2018-2022, authors of CMNE. All rights reserved.
# @license   MIT
# @brief     Methods to train the LSTM model
# ---------------------------------------------------------------------------

#%%
import os
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import datetime

from keras.models import Sequential
from keras.layers import Dense
from keras.layers import LSTM

from keras.callbacks import TensorBoard

from .settings import Settings
from .data import Data
from .data import generate_lstm_batches


def train(
    settings: Settings,
    data: Data,
    minibatch_size: int=30,
    steps_per_ep: int=25,
    num_epochs: int=100,
    lstm_look_back: int=80,
    num_unit: int=1280,
    idx: int=None,
    verbose: bool=False
) -> str:
    """
    Train the LSTM model.

    Args:
        settings: Settings object
        data: Data object
        minibatch_size: Size of the minibatches
        steps_per_ep: Number of steps per epoch
        num_epochs: Number of epochs
        lstm_look_back: Number of time steps to look back
        num_unit: Number of units in the LSTM layer
        idx: selection of the training epochs, if None all epochs are used
    
    Returns:
        model_name: Name of the trained model
    """    
    
    ###################################################################################################
    # The Script
    ###################################################################################################
    
    num_features_in = data.inv_op()['nsource']
    num_labels_out = num_features_in
    
    # TensorBoard Callback
    tbCallBack = TensorBoard(log_dir=settings.tb_log_dir(), histogram_freq=1, write_graph=True, write_images=True)

    #time_steps_in = lstm_look_back
    # create the Data Generator
    data_generator = generate_lstm_batches(epochs=data.train_epochs(idx=idx), inverse_operator=data.inv_op(), lambda2=data.lambda2(), method=data.method(), look_back=lstm_look_back, batch_size=minibatch_size)
    
    # create LSTM model
    model = None
    model = Sequential()
    model.add(LSTM(num_unit, activation='tanh', return_sequences=False, input_shape=(lstm_look_back,num_features_in)))
    model.add(Dense(num_labels_out, activation='linear'))
    
    # compile the model
    model.compile(loss = 'mean_squared_error', optimizer = 'adam')
    
    # Train - fit the model :D
    fitting_result = model.fit_generator(data_generator, steps_per_epoch = steps_per_ep, epochs = num_epochs, verbose=1, callbacks=[tbCallBack], validation_data=None, class_weight=None, workers=1)
    
    # # let's get some predictions
    # test_predict = model.predict(test_features)
    
    
    ###################################################################################################
    # Save Results
    ###################################################################################################
    
    date_stamp = datetime.datetime.now().strftime('%Y-%m-%d_%H%M%S')

    fname_model = settings.results_models_path() + '/model_' + settings.modality() + '_nu_' + str(num_unit) +'_lb_' + str(lstm_look_back) + '_' + date_stamp + '.h5'
    fname_training_loss = settings.results_training_path() + '/loss_' + settings.modality() + '_nu_' + str(num_unit) +'_lb_' + str(lstm_look_back) + '_' + date_stamp + '.txt'
    fname_result_fig = settings.results_img_path() + '/loss_' + settings.modality() + '_nu_' + str(num_unit) +'_lb_' + str(lstm_look_back) + '_' + date_stamp + '.png'
    
    history_losses = fitting_result.history['loss']
    
    # save model
    model.save(fname_model)
    
    # # plot the data
    if verbose:
        print('Testing Prediction',test_predict)
        print('Testing Reference',test_labels)
    
    # save loss
    np.savetxt(fname_training_loss, fitting_result.history['loss'])
    
    # save data plot
    plt.figure()
    plt.plot(fitting_result.history['loss'])
    plt.xlabel('Minibatch number')
    plt.ylabel('Loss')
    plt.title('Minibatch run vs. Training loss')
    #axes = plt.gca()
    #axes.set_xlim([xmin,xmax])
    #axes.set_ylim([0,1.2])
    fig = plt.gcf()
    fig.set_size_inches(8, 6)
    plt.savefig(fname_result_fig, dpi=300)
    
    # plot data
    if verbose:
        plt.show()

    return fname_model


"""Setup cmne."""

import sys

import os
import os.path as op

from setuptools import setup

# Give setuptools a hint to complain if it's too old a version
# 30.3.0 allows us to put most metadata in setup.cfg
# 38.3.0 contains most setup.cfg bugfixes
# Should match pyproject.toml
SETUP_REQUIRES = ["setuptools >= 38.3.0"]
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

# get the version (don't import mne here, so dependencies are not needed)
version = None
with open(op.join('cmne', '_version.py'), 'r') as fid:
    for line in (line.strip() for line in fid):
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\'')
            break
if version is None:
    raise RuntimeError('Could not determine version')


descr = """Contextual Minimum Norm Estimates (CMNE)"""

DISTNAME = 'cmne'
DESCRIPTION = descr
MAINTAINER = 'Christoph Dinh'
MAINTAINER_EMAIL = 'christoph.dinh@brain-link.de'
URL = 'https://github.com/chdinh/cmne'
LICENSE = 'MIT License'
DOWNLOAD_URL = 'https://github.com/chdinh/cmne'
VERSION = version


def package_tree(pkgroot):
    """Get the submodule list."""
    # Adapted from VisPy
    path = op.dirname(__file__)
    subdirs = [op.relpath(i[0], path).replace(op.sep, '.')
               for i in os.walk(op.join(path, pkgroot))
               if '__init__.py' in i[2]]
    return sorted(subdirs)


if __name__ == "__main__":
    if op.exists('MANIFEST'):
        os.remove('MANIFEST')

    with open('README.rst', 'r') as fid:
        long_description = fid.read()

    install_requires = list()
    with open('requirements.txt', 'r') as fid:
        for line in fid:
            req = line.strip()
            install_requires.append(req)
            
    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          include_package_data=True,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          long_description=long_description,
          long_description_content_type='text/x-rst',
          zip_safe=False,  # the package can run out of an .egg file
          classifiers=['Intended Audience :: Science/Research',
                       'Intended Audience :: Developers',
                       'License :: OSI Approved :: MIT License',
                       'Topic :: Scientific/Engineering',
                       'Programming Language :: Python :: 3',
                       'Development Status :: 1 - Planning',
                       'Operating System :: OS Independent',
                       ],
        keywords='MEG EEG spatiotemporal source estimation spatial filtering grid-based Markov localization LSTM deep learning',
        project_urls={
        'Source': 'https://github.com/chdinh/cmne/',
        'Tracker': 'https://github.com/chdinh/cmne/issues/',
        },
        platforms='any',
        python_requires='>=3.7',
        install_requires=install_requires,
        setup_requires=SETUP_REQUIRES,
        packages=package_tree('cmne'))


"""Contextual Minimum Norm Estimates (CMNE)"""

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.devN' where N is an integer.
#

from ._version import __version__

from .settings import (Settings)
from .data import (Data)
from .train import (train)